#!/usr/bin/python
# 0.0

import io
from sys import argv
from struct import unpack, pack

# open input
header_infile_name = argv[1]
header_infile = open(header_infile_name, 'rb')

data_infile_name = argv[2]
data_infile = open(data_infile_name, 'rb')

def process_sample(sample_header_offset, outfile_name):
    header_infile.seek(sample_header_offset+4)
    sample_rate = unpack('>I', header_infile.read(4))[0]

    header_infile.seek(sample_header_offset+0x10)
    true_header_offset = unpack('>I', header_infile.read(4))[0]
    
    header_infile.seek(true_header_offset)

    sample_size, sample_offset, info_offset, coef_offset = \
        unpack('>IIII', header_infile.read(16))

    # read general header
    header_infile.seek(info_offset)
    unk1, sample_count, unk2, unk3 = unpack('>IIII', header_infile.read(16))

    # read coefficient bank
    header_infile.seek(coef_offset)
    channels1, npredictors = unpack('>II', header_infile.read(8))

    coefs = {}
    for i in range(0,npredictors*16):
        coefs[i] = unpack('>h', header_infile.read(2))[0]

    print('decoding %s at %08x' % (outfile_name, sample_offset))
    outfile = start_RIFF(outfile_name, sample_count, sample_rate)
    decode_VADPCM(npredictors, coefs, sample_offset, sample_count, outfile)
    outfile.close()

def start_RIFF(outfile_name, sample_count, sample_rate):
    outfile = open(outfile_name,'wb')

    PCM_size = sample_count * 2
    RIFF_size = PCM_size + 0x2c

    # RIFF identification
    outfile.write("RIFF")
    outfile.write(pack('<I', RIFF_size-8))
    outfile.write("WAVE")

    # fmt chunk
    outfile.write("fmt ")
    outfile.write(pack('<I', 0x10))
    outfile.write(pack('<HHIIHH', 1, 1, sample_rate, sample_rate*2, 2, 16))

    # data chunk
    outfile.write("data")
    outfile.write(pack('<I', PCM_size))

    return outfile

# based on icemario's code as found in N64AIFCAudio.cpp
# clips a little...
def decode_VADPCM(npredictors, coefs, sample_offset, sample_count, outfile):
    data_infile.seek(sample_offset)

    clip_count = 0

    hist = (0,0,0,0,0,0,0,0)
    out = {}

    for i in range(0, sample_count, 16):
        frame = data_infile.read(9)

        scale = 1<<(ord(frame[0])>>4)
        pred = (ord(frame[0])&0xf) * 16

        for k in range(2):
            samples = {}
            for j in range(8):
                sample = ord(frame[1+k*4+j/2])
                if (j&1):
                    sample = sample&0xf
                else:
                    sample = sample>>4

                if sample >= 8:
                    sample -= 16

                samples[j] = sample * scale

            for j in range(8):
                total = coefs[pred+0+j] * hist[6]
                total += coefs[pred+8+j] * hist[7]

                if j>0:
                    for x in range(j):
                        total += samples[((j-1)-x)] * coefs[pred+8+x]

                total = ((samples[j] << 11) + total) >> 11

                if (total > 32767):
                    total = 32767
                    clip_count += 1
                elif total < -32768:
                    total = -32768
                    clip_count += 1

                outfile.write(pack('<h', total))

                out[j] = total
            hist = out
            out = {}

    if clip_count > 0:
        print("clipped %d times" % clip_count)

    outfile.close()

for sample_idx in range(0,128):
#for sample_idx in range(2,3):
    header_infile.seek(sample_idx * 4)
    sample_offset = unpack('>I', header_infile.read(4))[0]

    if sample_offset == 0:
        continue

    #print('Sample %02x: %08x' % (sample_idx, sample_offset))
    outfile_name = 'sample_%02x.wav' % sample_idx
    process_sample(sample_offset, outfile_name)
